/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2006
*/

#include "ScatterWeb.System.h"

config_t Configuration; 

UINT8 _Configuration_chooseID() {
    INT8 vn0 = *(INT8*)0x107F;
    INT8 vn1 = *(INT8*)0x10FF;
    if (vn0 == -1) return 1;
    else if (vn1 == -1) return 0;
    return ( (INT8)(vn0 - vn1) < 0 );
}

bool _Configuration_writeID(UINT16 addr, UINT8 buf) {
	void* taddr = (void*) addr;
	void* tbuf = (void*) &buf;
    volatile INT8 *newp;
    INT8 *oldp;
    INT8 *bufp = (INT8*) tbuf;
    INT8 ver;
    UINT16 i;
    if (0x7E + 2 < (UINT16) taddr + 1) return False;
    taddr += 0x1000;
    newp = oldp = (INT8*) 0x1000;
    if (_Configuration_chooseID()) oldp += 128;
    else {
      taddr += 128;
      newp += 128;
    }
	dint();
    FCTL2 = FWKEY + FSSEL1 + FN2;
    FCTL3 = FWKEY;
    FCTL1 = FWKEY + ERASE;
    *newp = 0;
    FCTL1 = FWKEY + WRT;
    for ( i = 0; i < 128-1; i++, newp++, oldp++ ) {
      if ((UINT16) newp < (UINT16) taddr || (UINT16) taddr+1 <= (UINT16) newp) *newp = *oldp;
	  else *newp = *bufp++;
	}
    ver = *oldp + 1;
    if (ver == -1) ver++;
    *newp = ver;
    FCTL1 = FWKEY;
    FCTL3 = FWKEY + LOCK;
	eint();
    return True;
}

UINT8 _Configuration_readID(UINT16 addr) {
    if (_Configuration_chooseID()) addr += 128;
    return *(UINT8*)addr;
}

// rcf <br> Prints complete configuration.
COMMAND(rcf, SERIALONLY) {
  Configuration_print();
}

// rid <br> Prints ID of the sensor board.
COMMAND(rid, 0) {
  String_write(term_reply, "[rid] %i", Configuration.id);
}

// sid nnnn <br> Sets ID of the sensor board. Value range: 0000-9999, Ex: sid 4711
COMMAND(sid, 0) {
  UINT16 temp = String_parseInt(&str[4], NULL);
  Configuration.id = temp;
  UINT8 t2 = temp;
  temp = temp >> 8;
  UINT8 t1 = temp ;
  _Configuration_writeID(80, t1);
  _Configuration_writeID(81, t2);
}

// rlo <br> Read log level, see ScatterWeb.Configuration.h for possible values.
COMMAND(rlo, 0) { 
  UINT8 l = Comm_logLevel & 0x0F;
  String_write(term_reply, "[rlo] %b [%s%s%s]", Comm_logLevel,
                      ((Comm_logLevel & DEBUG_RADIO) ? "RADIO " : ""),
                      ((Comm_logLevel & DEBUG_DCO) ? "DCO " : ""),
                      ( l==0 ? "DISABLED" : ( l==NO ? "NO" : ( l==LOW ? "LOW" : ( l==MEDIUM ? "MEDIUM" : ( l==HIGH ? "HIGH" : "VERBOSE ")))) )
  ); 
}
    
// slo xx <br> Set log level, see ScatterWeb.Configuration.h for possible values.
COMMAND(slo, 0) {
  Comm_logLevel = String_fromHexToByte(&str[4]); 
}

// Read boot log level, see ScatterWeb.Configuration.h for possible values.
COMMAND(rlb, 0) {   
  UINT8 l = Configuration.boot_logLevel & 0x0F;
  String_write(term_reply, "[slb] %b [%s%s%s]", Configuration.boot_logLevel,
                      ((Configuration.boot_logLevel & DEBUG_RADIO) ? "RADIO " : ""),
                      ((Configuration.boot_logLevel & DEBUG_DCO) ? "DCO " : ""),
                      ( l==0 ? "DISABLED" : ( l==NO ? "NO" : ( l==LOW ? "LOW" : ( l==MEDIUM ? "MEDIUM" : ( l==HIGH ? "HIGH" : "VERBOSE ")))) )
  );  
}
               
// Set boot log level, see ScatterWeb.Configuration.h for possible values.
COMMAND(slb, 0) { 
  Configuration.boot_logLevel = String_fromHexToByte( &str[4]);
}

// rff <br> Reads firmware flags, see ScatterWeb.Configuration.h for possible values.
COMMAND(rff, 0) { 
 String_write(term_reply, "[rff] %b [%s%s]", Configuration.firmwareFlags,
                          ((Configuration.firmwareFlags & FF_PROGRAMMABLE) ? "PROG " : ""),
                          ((Configuration.firmwareFlags & FF_DCOCHECKER) ? "DCO " : "")
  );
}

// sff xx <br> Sets firmware flags, see ScatterWeb.Configuration.h for possible values.
COMMAND(sff, 0) { 
  Configuration.firmwareFlags = String_fromHexToByte( &str[4]); 
}

bool Configuration_init() {
  Configuration.id = (_Configuration_readID(0x1050) << 8) + _Configuration_readID(0x1051);
  Configuration.transmitPower = CC1020_paPower;  
  Configuration.boot_logLevel = 1;
  Configuration.firmwareFlags = 0;
  CC1020_loadSettings();
  Comm_logLevel = Configuration.boot_logLevel;
  return True;
}

bool Configuration_print() {
  Comm_print("\r\n+==============================================================\r\n");
  Comm_print("| DCO: %b BCS: %b\r\n", DCOCTL, BCSCTL1);  
  Comm_print("| Id: %i\r\n", Configuration.id);
  Comm_print("| transmitPower: %i  \r\n", Configuration.transmitPower);
  term_rlo(NULL); Comm_print("| logLevel:      %s", term_reply);
  term_rlb(NULL); Comm_print(" (%s)\r\n", term_reply);
  term_rff(NULL); Comm_print("| firmwareFlags: %s\r\n", term_reply);
  Comm_print("| FLASH[EOS]: %D\r\n", &firmwareData);
  Comm_print("| FLASH[APP]: %D\r\n", firmwareData.end);
  Comm_print("+==============================================================\r\n");
  term_reply[0] = '\0';
  return True;
}
