/*
Copyright 2006, Freie Universitaet Berlin. All rights reserved.

These sources were developed at the Freie Universitt Berlin, Computer
Systems and Telematics group.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.
 
- Neither the name of Freie Universitaet Berlin (FUB) nor the names of its
contributors may be used to endorse or promote products derived from
this software without specific prior written permission.

This software is provided by FUB and the contributors on an "as is"
basis, without any representations or warranties of any kind, express
or implied including, but not limited to, representations or
warranties of non-infringement, merchantability or fitness for a
particular purpose. In no event shall FUB or contributors be liable
for any direct, indirect, incidental, special, exemplary, or
consequential damages (including, but not limited to, procurement of
substitute goods or services; loss of use, data, or profits; or
business interruption) however caused and on any theory of liability,
whether in contract, strict liability, or tort (including negligence
or otherwise) arising in any way out of the use of this software, even
if advised of the possibility of such damage.

This implementation was developed by the CST group at the FUB.

For documentation and questions please use the web site
http://scatterweb.mi.fu-berlin.de and the mailinglist
scatterweb@lists.spline.inf.fu-berlin.de (subscription via the Website).
Berlin, 2006
*/

#include "ScatterWeb.System.h"

#define startWatchdog() do { WDTCTL = WDT_ARST_1000; }while(0)
#define stopWatchdog()  do { WDTCTL = WDTPW + WDTHOLD; }while(0)
#define ISR __attribute__((naked)) __attribute__((section(".isr")))
#define DELTA 1100                   // Target DCO = DELTA(1100)*(4096Hz) , DELTA=550 => 4505600 Hz  */
#define UP 1                         // Temporary.
#define DOWN 2                       // Temporary.
#define DCOLOW  (1)                  // Symbolic name, DCO was last measure too low.
#define DCOHIGH (2)                  // Symbolic name, DCO was last measure too high.
#define DCODIFF (3)                  // Difference of dcoDiff to DELTA when DCO measure is taken as too high/low.
#define RESET (WDTCTL = 0x0000)
#define WDSTATE (*((UINT16*)0x09fe)) // 2 bytes are reserved at main
#define LOADER         __attribute__((section(".loader")))

LOADER __attribute__ ((naked)) void _reset_vector__() {
  __asm__ __volatile__("dint"::);
  __asm__ __volatile__("br #main"::); // Branch to main.
}

__attribute__((section(".ivt"))) void* volatile IVT[16];

volatile UINT16 dcoCap;        // Contains always last TBCCR6 Value (Capture on ACLK)
volatile UINT16 dcoDiff;       // Evreytime dcoDiff==0 TIMER_B1 interrupt writes a new DCO diff measure into it.
volatile UINT8  checkDCOState; // DCO will only be changed if two times low or high in a row to avoid peaks. Implemented with this variable.
volatile UINT16 runModule;
bool isWatchdogOn;

fp_vp callbacks[NUMBER_CALLBACKS]; // The callbacks set via System_registerCallback are stored here.

extern void Process_init(); // The only function in userapp that can be called from firmware (except callbacks).

// software wd
interrupt(NOVECTOR) wdt_timera1() { }

// DCO Checker. Interrupt occurs every ~250us. Continously updates ::dcoDiff.
interrupt(NOVECTOR) Timer_B1() {
  dcoDiff = TBCCR6 - dcoCap;
  dcoCap = TBCCR6;            
  TBCCTL6 &= ~(CCIFG);
}

interrupt(NOVECTOR) undefined_isr() {
  Comm_print("UNDEFINED ISR!!!!\n");
  WDTCTL = 0xDEAD;
}

interrupt(NOVECTOR) undefined_isr0() { Comm_print("##0!!\n"); WDTCTL = 0xDEAD; }   //  0 0xFFE0 DAC12/DMA
interrupt(NOVECTOR) undefined_isr1() { Comm_print("##1!!\n"); WDTCTL = 0xDEAD; }   //  2 0xFFE2 Port 2
interrupt(NOVECTOR) undefined_isr2() { Comm_print("##2!!\n"); WDTCTL = 0xDEAD; }   //  4 0xFFF4 USART 1 Transmit
interrupt(NOVECTOR) undefined_isr3() { Comm_print("##3!!\n"); WDTCTL = 0xDEAD; }   //  6 0xFFF6 USART 1 Receive
interrupt(NOVECTOR) undefined_isr4() { Comm_print("##4!!\n"); WDTCTL = 0xDEAD; }   //  8 0xFFE8 Port 1
interrupt(NOVECTOR) undefined_isr5() { Comm_print("##5!!\n"); WDTCTL = 0xDEAD; }   // 10 0xFFEA Timer A CC1-2, TA
interrupt(NOVECTOR) undefined_isr6() { Comm_print("##6!!\n"); WDTCTL = 0xDEAD; }   // 12 0xFFEC Timer A CC0
interrupt(NOVECTOR) undefined_isr7() { Comm_print("##7!!\n"); WDTCTL = 0xDEAD; }   // 14 0xFFEE ADC
interrupt(NOVECTOR) undefined_isr8() { Comm_print("##8!!\n"); WDTCTL = 0xDEAD; }   // 16 0xFFF0 USART 0 Transmit
interrupt(NOVECTOR) undefined_isr9() { Comm_print("##9!!\n"); WDTCTL = 0xDEAD; }   // 18 0xFFF2 USART 0 Receive, I2C tx/rx
interrupt(NOVECTOR) undefined_isr10() { Comm_print("##10!!\n"); WDTCTL = 0xDEAD; } // 20 0xFFF4 Watchdog Timer 
interrupt(NOVECTOR) undefined_isr11() { Comm_print("##11!!\n"); WDTCTL = 0xDEAD; } // 22 0xFFF6 Comparator A 
interrupt(NOVECTOR) undefined_isr12() { Comm_print("##12!!\n"); WDTCTL = 0xDEAD; } // 24 0xFFF8 Timer B CC1-2, TB 
interrupt(NOVECTOR) undefined_isr13() { Comm_print("##13!!\n"); WDTCTL = 0xDEAD; } // 26 0xFFFA Timer B CC0 
interrupt(NOVECTOR) undefined_isr14() { Comm_print("##14!!\n"); WDTCTL = 0xDEAD; } // 28 0xFFFC Non-maskable 
interrupt(NOVECTOR) undefined_isr15() { Comm_print("##15!!\n"); WDTCTL = 0xDEAD; } // reset, not overwritten in flash end

ISR interrupt(DACDMA_VECTOR)     dacdma_isr() { __asm__ __volatile__("br &0x1100"::); } //  0 0xFFE0 DAC12/DMA 
ISR interrupt(PORT2_VECTOR)       port2_isr() { __asm__ __volatile__("br &0x1102"::); } //  2 0xFFE2 Port 2 
ISR interrupt(USART1TX_VECTOR)     u1tx_isr() { __asm__ __volatile__("br &0x1104"::); } //  4 0xFFF4 USART 1 Transmit 
ISR interrupt(USART1RX_VECTOR)     u1rx_isr() { __asm__ __volatile__("br &0x1106"::); } //  6 0xFFF6 USART 1 Receive 
ISR interrupt(TIMERA1_VECTOR)   timera1_isr() { __asm__ __volatile__("br &0x110A"::); } // 10 0xFFEA Timer A CC1-2, TA 
ISR interrupt(TIMERA0_VECTOR)   timera0_isr() { __asm__ __volatile__("br &0x110C"::); } // 12 0xFFEC Timer A CC0 
ISR interrupt(ADC_VECTOR)           adc_isr() { __asm__ __volatile__("br &0x110E"::); } // 14 0xFFEE ADC 
ISR interrupt(USART0TX_VECTOR)     u0tx_isr() { __asm__ __volatile__("br &0x1110"::); } // 16 0xFFF0 USART 0 Transmit 
ISR interrupt(USART0RX_VECTOR)     u0rx_isr() { __asm__ __volatile__("br &0x1112"::); } // 18 0xFFF2 USART 0 Receive, I2C tx/rx 
ISR interrupt(WDT_VECTOR)           wdt_isr() { __asm__ __volatile__("br &0x1114"::); } // 20 0xFFF4 Watchdog Timer 
ISR interrupt(COMPARATORA_VECTOR) compa_isr() { __asm__ __volatile__("br &0x1116"::); } // 22 0xFFF6 Comparator A 
ISR interrupt(TIMERB1_VECTOR)   timerb1_isr() { __asm__ __volatile__("br &0x1118"::); } // 24 0xFFF8 Timer B CC1-2, TB 
ISR interrupt(TIMERB0_VECTOR)   timerb0_isr() { __asm__ __volatile__("br &0x111A"::); } // 26 0xFFFA Timer B CC0 
ISR interrupt(NMI_VECTOR)           nmi_isr() { __asm__ __volatile__("br &0x111C"::); } // 28 0xFFFC Non-maskable 

// Subroutine to check and readjust the dco.
//
// This routine will run whenevery the Timer_B1 interrupt has left a new value in dcoDiff.
// The algorithm checks if the measured frequency is more than DCODIFF steps off the
// desired DELTA frequency. If that is the case two times in row the algorithm
// lowers/highers the dco frequency. Set bit DEBUG_DCO in Configuration.debugFlags to see
// readjustments of the dco.
//
// If stable, diff will always oscillate between two numbers, for example 1097 and 1098.
// At theses frequency (~4,5 Mhz) one step of DCO means 4 measured diffDCO steps,
// for example one DCO step up with the previous example will produce 1101/1102.
// To avoid constant changing of DCO frequency and proviode some hysteresis DCODIFF
// should be at least 3 (19200 rf and 115200 baud serial comm get only lost at diffs greater 50).
void _System_checkDCO() {
  if(dcoDiff<(DELTA-DCODIFF)) {      // frequency too low
    if(checkDCOState != DCOLOW) checkDCOState = DCOLOW;
    else {
      if(Comm_logLevel & DEBUG_DCO) Comm_print("DCO++");
      DCOCTL++;
      if (DCOCTL == 0x00) {
        if((BCSCTL1 & 0x07) == 0x07) {
          if(Comm_logLevel & DEBUG_DCO) Comm_print("Frequency too high!!!\r\n");
          WDTCTL = 0xDEAD;
        }
        BCSCTL1++;      // Did DCO role over? // -> Select next higher RSEL
      }
      if(Comm_logLevel & DEBUG_DCO) Comm_print(" %b %b >%i\r\n", DCOCTL, BCSCTL1, dcoDiff);
    }
  }
  else {
    if(dcoDiff>(DELTA+DCODIFF)) {      // frequency too high
      if(checkDCOState != DCOHIGH) {
        checkDCOState = DCOHIGH;
      }
      else {
        if(Comm_logLevel & DEBUG_DCO) Comm_print("DCO--");
        DCOCTL--;
        if (DCOCTL == 0xFF) {
          if(!((BCSCTL1 & 0x07) == 0x00)) {
            if(Comm_logLevel & DEBUG_DCO) Comm_print("Frequency too low!!!\r\n");
            WDTCTL = 0xDEAD;
          }
          BCSCTL1--;      // Did DCO role under? -> Select next lower RSEL
        }
        if(Comm_logLevel & DEBUG_DCO) Comm_print(" %b %b >%i\r\n", DCOCTL, BCSCTL1, dcoDiff);
      }
    }
    else checkDCOState = 0; // freqency was ok
  }
  Timers_add(100, _System_checkDCO, 0xFFFF);
}

// If the A/D-Converter is enabled via _System_enableADC, the values of the analog
// inputs are continously converted
void _System_initADC() {
  //Setup ADC12, ref., sampling time, REFON for DAC!!
  ADC12CTL0 = SHT0_15+SHT1_15+MSC;  
  ADC12CTL1 = SHP + CONSEQ_3 + CSTARTADD_0;   // Use sampling timer, repeat-sequence-of-channels
  ADC12MCTL0 = (INCH_0 + SREF_0);             // P60: AIN0
  ADC12MCTL1 = (INCH_0 + SREF_0);             // P60: AIN0
  ADC12MCTL2 = (INCH_0 + SREF_0);             // P60: AIN0
  ADC12MCTL3 = (INCH_1 + SREF_0);             // P61: AIN1
  ADC12MCTL4 = (INCH_1 + SREF_0);             // P61: AIN1
  ADC12MCTL5 = (INCH_1 + SREF_0);             // P61: AIN1
  ADC12MCTL6 = (INCH_2 + SREF_0);             // P62: AIN2
  ADC12MCTL7 = (INCH_2 + SREF_0);             // P62: AIN2
  ADC12MCTL8 = (INCH_2 + SREF_0);             // P62: AIN2
  ADC12MCTL9 = (INCH_2 + SREF_0);             // P62: AIN2
  ADC12MCTL10 = (INCH_3 + SREF_0);            // P63: AIN2
  ADC12MCTL11 = (INCH_4 + SREF_0);            // P64: AIN3
  ADC12MCTL12 = (INCH_5 + SREF_0);            // P65: AIN4
  ADC12MCTL13 = (INCH_10 + SREF_0) + EOS;     // Temp diode
}

// Performs initialization of all ports.
// Sets SELECTION, DIRECTION and OUT registers for P1 to P6, see code for details.
void _System_initPorts() {
  P1SEL = 0x00;
  P1OUT = 0x00;
  P1DIR = 0x00;
  P2SEL = 0x00;
  P2OUT = 0x18;
  P2DIR = 0x1A;
  P3SEL = 0x00;
  P3OUT = 0x09;
  P3DIR = 0x21;
  P4SEL = 0x00;
  P4OUT = 0x00;
  P4DIR = 0x00;
  P5SEL = 0x00;
  P5OUT = 0xF9;
  P5DIR = 0xFB;
  P6SEL = 0x07;
  P6OUT = 0x00;
  P6DIR = 0xC8;
}

void _System_initIVTtable() {
  UINT8 i;
  for(i=0; i<16; i++) IVT[i] = undefined_isr;
  IVT[0] = undefined_isr0;
  IVT[1] = undefined_isr1;
  IVT[2] = undefined_isr2;
  IVT[3] = undefined_isr3;
  IVT[4] = undefined_isr4;
  IVT[5] = undefined_isr5;
  IVT[6] = undefined_isr6;
  IVT[7] = undefined_isr7;
  IVT[8] = undefined_isr8;
  IVT[9] = undefined_isr9;
  IVT[10] = undefined_isr10;
  IVT[11] = undefined_isr11;
  IVT[12] = undefined_isr12;
  IVT[13] = undefined_isr13;
  IVT[14] = undefined_isr14;
  IVT[15] = undefined_isr15; 
}

// Enables and starts ADC.
// Enables the A/D-Converter, uses ~1 mA.
bool System_enableADC() {
  _System_initADC();
  ADC12CTL0 |= ADC12ON;
  System_wait(600);     // Delay for reference start-up
  ADC12CTL0 |= ENC;     // enable conversion
  ADC12CTL0 |= ADC12SC; // Start conversion
  return True;
}

// Stops and disables ADC.
// Disables the A/D-Converter.
bool System_disableADC() {
  ADC12CTL0 &= ~ENC;       // Stop conversion
  System_wait(600);        // Wait, TODO exchange with test for ADC12BUSY &test
  ADC12CTL0 &= ~(ADC12ON); // turn off ADC12
  return True;
}

void System_reset() { 
  FCTL3 = 0xDEAD;
}

bool System_registerCallback(UINT8 type, fp_t fp) {
  callbacks[type] = fp;
  return True;
}

bool System_delay(unsigned int ms) {
  UINT16 tar = TAR;
  ms <<= 2;     // Wait with ACLK, take ms * 4 to get roughly ms (1024 hz)
  while(TAR - tar < ms) _NOP();
  return True;
}

bool System_wait(UINT32 w) {
  volatile UINT32 i; 
  for(i=0;i<w;i++) nop(); 
  return True;
}

bool System_setDCO() {
  UINT16 Compare, Oldcapture = 0, i, direction = 0;
  // ACLK is devided by 8 = 4096 Hz. RSEL=6 no division for MCLK and SSMCLK. XT2 is off.
  BCSCTL1 = 0xB7;
  // Init FLL to desired frequency using the 32762Hz crystal DCO frquenzy = 2,4576 MHz
  BCSCTL2 = 0x00;    
  DCOCTL = 0xFF;
  for (i = 0xFFFF; i > 0; i--);     // Delay for XTAL to settle
  CCTL2 = CCIS0 + CM0 + CAP;        // Define CCR2, CAP, ACLK
  TACTL = TASSEL1 + TACLR + MC1;    // SMCLK, continous mode
  while (1) {
    CCTL2 &= ~(CCIFG + COV);        // clear IFG and possible overflow
                                    // 1. capture
    while (!(CCTL2 & CCIFG));               // Wait until capture occured!
    CCTL2 &= ~CCIFG;                        // Capture occured, clear flag
    Oldcapture = CCR2;                      // Save current captured SMCLK
                                    // 2. - 5. capture, "software divider" so aclk stays
                                    // untouched and 2400 bps are working!
      while (!(CCTL2 & CCIFG));             // Wait until capture occured!
      CCTL2 &= ~CCIFG;                      // Capture occured, clear flag
      Compare = CCR2;
    Compare = Compare - Oldcapture;         // SMCLK difference
    if (DELTA == Compare) break;            // if equal, leave "while(1)"
    else {
      if (DELTA < Compare) {             // DCO is too fast, slow it down
        if(direction == UP) break;       // stop if we were going up (jumped over exact delta)
        if(!direction) direction = DOWN; // if starting set direction
        DCOCTL--;
        if (DCOCTL == 0xFF) {
          if(!((BCSCTL1 & 0x07) == 0x00)) WDTCTL = 0xDEAD;
          BCSCTL1--;            // Did DCO role under? -> Select next lower RSEL
        }
      }
      else {
        if(direction == DOWN) break;   // stop if we were going down (jumped over exact delta)
        if(!direction) direction = UP; // if starting set direction
        DCOCTL++;
        if (DCOCTL == 0x00) {
          if((BCSCTL1 & 0x07) == 0x07) WDTCTL = 0xDEAD;
          BCSCTL1++;            // Did DCO role over? // -> Select next higher RSEL
        }
      }
    }
  }  
  CCTL2 = 0;      // Stop CCR2 function
  TACTL = 0;      // Stop Timer_A
  dcoDiff = 0;
  checkDCOState = 0;
  return True;
}

bool System_startWatchdog() {
  isWatchdogOn = True;
  startWatchdog();
  return True;
}

bool System_stopWatchdog() {
  isWatchdogOn = False;
  stopWatchdog();
  return True;
}

// Prints stack pointer.
COMMAND(rsp, 0) {
  String_write(term_reply, "[rsp] %x", READ_SP);
}

//  endless loop, calling callbacks and doing internal stuff.
interrupt(NOVECTOR) superloop() {
  loopstart:
    CACTL1 &= ~CAIE;
    eint();
	if(isWatchdogOn) startWatchdog();
	if(runModule & MF_TIMER) Timers_eventHandler();  
	if(runModule & MF_RADIO_RX) Net_rxHandler();
	if(runModule & MF_RADIO_TX) Net_txHandler();
    if(runModule & MF_SERIAL) {
      extern volatile UINT8* serial_line;
      if(serial_line != 0) {
      	if(callbacks[C_SERIAL]) callbacks[C_SERIAL]((void*)serial_line);
      }
    }
	if(runModule & MF_SENSOR) {
      if(callbacks[C_SENSOR]) callbacks[C_SENSOR](&Data_sensorEventValue);
	  runModule &= ~MF_SENSOR;
    }
    dint(); _NOP();
    if(runModule==0) {
      stopWatchdog();
      CACTL1 = CAIE; // clears also ifg
      return;
    } 
    else {  
      goto loopstart;
    }    
}

// Runs the bootsequence
__attribute__((reserve(2))) 
__attribute__((section(".fw_main"))) int main() {
  extern volatile UINT16 irqLock;
  dint();
  DMA0CTL &= ~(DMAIFG + DMAIE);
  irqLock = 0;
  WDTCTL = WDTPW + WDTHOLD;
  _System_initPorts();
  CACTL1 = 0x0000;
  _System_initIVTtable();
  CC1020_reset();
  CC1020_setupPD();
  runModule = 0;   
  System_enableADC();
  System_setDCO();
  Configuration_init();
  Timers_init();
  eint(); 
  _NOP();
  Comm_init(BPS_115200);
  //Current time is always set to 9:37
  time_t time = { 34620, 0 };
  Time_set(&time);
  Net_init();
  Net_on(TXMODE_BACKOFF);
  Data_init(0);
  Configuration_print();
  volatile UINT16 i;
  for(i=0; i<NUMBER_CALLBACKS; i++) callbacks[i] = 0;
  isWatchdogOn = True;
  Messaging_on();
  Process_init();
  SET_ISR(COMPARATORA_VECTOR, superloop);
  _NOP(); _NOP();
  CACTL1 = CAIE;
  sleepforever:
    LPM1;
    goto sleepforever;
}

__attribute__((section(".fw_header"))) const imageData_t firmwareData = {
   0x4713,               // Key
   &__fw_start,          // Start
   &__fw_end,            // End  
   0x0000,               // Type
   "MSBEOS 1.00",        // Must be 20 chars max (with \0 21 alltogether).
   __DATE__ " " __TIME__ // Only filled when programming.o is rebuilt!
};
